import torch
import pytest
from goldener.vectorize import GoldVectorizer, Filter2DWithCount, FilterLocation


class TestGoldVectorizer:
    def make_tensor(self, shape=(2, 5, 2)):
        return torch.randint(0, 100, shape)

    def test_vectorize_no_y(self):
        x = self.make_tensor()
        v = GoldVectorizer()
        vec = v.vectorize(x)
        assert vec.vectors.shape == (4, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 0, 1, 1]))

    def test_vectorize_with_y(self):
        x = self.make_tensor()
        y = torch.ones(2, 1, 2)
        y[0, 0, 0] = 0
        v = GoldVectorizer()
        vec = v.vectorize(x, y)
        assert vec.vectors.shape == (3, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 1, 1]))

    def test_vectorize_with_keep(self):
        x = self.make_tensor()
        keep = Filter2DWithCount(
            filter_count=1, filter_location=FilterLocation.START, keep=True
        )
        v = GoldVectorizer(keep=keep)
        vec = v.vectorize(x)
        assert vec.vectors.shape == (2, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 1]))

    def test_vectorize_with_remove(self):
        x = self.make_tensor()
        remove = Filter2DWithCount(
            filter_count=1, filter_location=FilterLocation.END, keep=False
        )
        v = GoldVectorizer(remove=remove)
        vec = v.vectorize(x)
        assert vec.vectors.shape == (2, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 1]))

    def test_vectorize_with_keep_and_remove(self):
        x = self.make_tensor((2, 5, 3))
        keep = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.START, keep=True
        )
        remove = Filter2DWithCount(
            filter_count=1, filter_location=FilterLocation.END, keep=False
        )
        v = GoldVectorizer(keep=keep, remove=remove)
        vec = v.vectorize(x)
        assert vec.vectors.shape == (2, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 1]))

    def test_vectorize_with_transform_y(self):
        x = self.make_tensor()
        shape = x.shape
        y = 10 * torch.ones((shape[0], 1, shape[2]))
        y[0, 0, 0] = 3
        y[1, 0, 0] = 3

        def transform_y(y):
            # Only keep rows where y > 5
            return (y > 5).to(torch.int64)

        v = GoldVectorizer(transform_y=transform_y)
        vec = v.vectorize(x, y)
        assert vec.vectors.shape == (2, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 1]))

    def test_vectorize_shape_mismatch(self):
        x = self.make_tensor()
        y = torch.ones(2, 1, 3)
        v = GoldVectorizer()
        with pytest.raises(ValueError):
            v.vectorize(x, y)

    def test_vectorize_2d_input(self):
        x = self.make_tensor((4, 5))
        v = GoldVectorizer()
        with pytest.raises(ValueError):
            v.vectorize(x)

    def test_vectorizer_invalid_keep_type(self):
        with pytest.raises(ValueError):
            GoldVectorizer(keep=Filter2DWithCount())

    def test_vectorizer_invalid_remove_type(self):
        with pytest.raises(ValueError):
            GoldVectorizer(remove=Filter2DWithCount())

    def test_vectorizer_invalid_random_type(self):
        with pytest.raises(ValueError):
            GoldVectorizer(
                random_filter=Filter2DWithCount(filter_location=FilterLocation.START)
            )


class TestFilter2DWithCount:
    def make_tensor(self):
        # 5x3 tensor with unique values for easy row checking
        return torch.arange(15).reshape(5, 3)

    def test_filter_start_keep(self):
        x = self.make_tensor()
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.START, keep=True
        )
        out = f.filter(x)
        assert out.shape[0] == 2
        assert torch.equal(out, x[:2])

    def test_filter_start_remove(self):
        x = self.make_tensor()
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.START, keep=False
        )
        out = f.filter(x)
        assert out.shape[0] == 3
        assert torch.equal(out, x[2:])

    def test_filter_end_keep(self):
        x = self.make_tensor()
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.END, keep=True
        )
        out = f.filter(x)
        assert out.shape[0] == 2
        assert torch.equal(out, x[-2:])

    def test_filter_end_remove(self):
        x = self.make_tensor()
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.END, keep=False
        )
        out = f.filter(x)
        assert out.shape[0] == 3
        assert torch.equal(out, x[:-2])

    def test_filter_random_keep(self):
        x = self.make_tensor()
        generator = torch.Generator().manual_seed(42)
        f = Filter2DWithCount(
            filter_count=2,
            filter_location=FilterLocation.RANDOM,
            keep=True,
            generator=generator,
        )
        out = f.filter(x)
        assert out.shape[0] == 2
        for row in out:
            assert any(torch.equal(row, r) for r in x)

    def test_filter_random_remove(self):
        x = self.make_tensor()
        generator = torch.Generator().manual_seed(42)
        f = Filter2DWithCount(
            filter_count=2,
            filter_location=FilterLocation.RANDOM,
            keep=False,
            generator=generator,
        )
        out = f.filter(x)
        assert out.shape[0] == 3
        for row in out:
            assert any(torch.equal(row, r) for r in x)

    def test_filter_tensor_dict(self):
        x = self.make_tensor()
        d = {"a": x.clone(), "b": x.clone() + 100}
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.START, keep=True
        )
        out = f.filter_tensors(d)
        assert isinstance(out, dict)
        assert set(out.keys()) == {"a", "b"}
        assert torch.equal(out["a"], x[:2])
        assert torch.equal(out["b"], x[:2] + 100)

    def test_filter_random_keep_tensor_dict(self):
        x = self.make_tensor()
        d = {"a": x.clone(), "b": x.clone()}
        generator = torch.Generator().manual_seed(123)
        f = Filter2DWithCount(
            filter_count=2,
            filter_location=FilterLocation.RANDOM,
            keep=True,
            generator=generator,
        )
        out = f.filter_tensors(d)
        assert isinstance(out, dict)
        assert set(out.keys()) == {"a", "b"}
        for tensor in out.values():
            assert tensor.shape[0] == 2
            for row in tensor:
                assert any(torch.equal(row, r) for r in x)

    def test_invalid_filter_count(self):
        with pytest.raises(ValueError):
            Filter2DWithCount(filter_count=0)

    def test_non_2d_input(self):
        x = torch.arange(10)
        f = Filter2DWithCount(filter_count=1)
        with pytest.raises(ValueError):
            f.filter(x)
        d = {"a": torch.arange(10)}
        with pytest.raises(ValueError):
            f.filter_tensors(d)

    def test_inconsistent_batch_size_dict(self):
        x = self.make_tensor()
        d = {"a": x, "b": x[:3]}
        f = Filter2DWithCount(filter_count=2)
        with pytest.raises(ValueError):
            f.filter_tensors(d)

    def test_filter_count_greater_than_rows(self):
        x = self.make_tensor()
        # filter_count > number of rows
        f = Filter2DWithCount(
            filter_count=10, filter_location=FilterLocation.START, keep=True
        )
        out = f.filter(x)
        assert (out == x).all()

    def test_dict_output_keys_and_shapes(self):
        x = self.make_tensor()
        d = {"a": x, "b": x + 1}
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.START, keep=True
        )
        out = f.filter_tensors(d)
        assert set(out.keys()) == {"a", "b"}
        assert out["a"].shape == (2, 3)
        assert out["b"].shape == (2, 3)



from dataclasses import dataclass

import pixeltable as pxt
from pixeltable.catalog import Table
from pixeltable.exprs import Expr
from torch.utils.data import Dataset

from goldener.describe import GoldDescriptor
from goldener.pxt_utils import (
    get_expr_from_column_name,
    pxt_torch_dataset_collate_fn,
    GoldPxtTorchDataset,
    get_array_column_shapes,
    _get_column_distinct_ratios,
    set_value_to_idx_rows,
)
from goldener.select import GoldSelector
from goldener.torch_utils import ResetableTorchIterableDataset
from goldener.utils import get_ratio_list_sum


@dataclass
class GoldSet:
    """Configuration for a gold set used for splitting.

    Attributes:
        name: Name of the gold set.
        ratio: Ratio of samples to assign to this set (between 0 and 1). This value
        cannot be one of 0 or 1 (goal is to select a subset of the full dataset).
    """

    name: str
    ratio: float

    def __post_init__(self) -> None:
        if not (0 < self.ratio < 1):
            raise ValueError("Ratio must be between 0 and 1.")


class GoldSplitter:
    """Splitter that divides a dataset into multiple sets based on features.

    When the sum of ratios of the provided sets during initialization is less than 1,
    an additional set named "not assigned" is added to cover the remaining samples.

    Attributes:
        sets: List of GoldSet configurations defining the splits.
        descriptor: GoldDescriptor used to describe the dataset.
        selector: GoldSelector used to select samples for each set. The collate_fn of the selector
        will be set to `pxt_torch_dataset_collate_fn`.
        class_key: Optional key for class-based stratification.
        drop_table: Whether to drop the described table after splitting.
    """

    def __init__(
        self,
        sets: list[GoldSet],
        descriptor: GoldDescriptor,
        selector: GoldSelector,
        class_key: str | None = None,
        drop_table: bool = False,
    ) -> None:
        """Initialize the GoldSplitter.

        Args:
            sets: List of GoldSet configurations defining the splits. When the sum of ratios is less than 1,
            an additional set named "not assigned" will be created to cover the remaining samples.
            descriptor: GoldDescriptor used to describe the dataset.
            selector: GoldSelector used to select samples for each set. The collate_fn of the selector
            will be set to `pxt_torch_dataset_collate_fn`.
            class_key: Optional key for class-based stratification.
            drop_table: Whether to drop the described table after splitting.

        Raises:
            ValueError: If set names are not unique or ratios do not sum to 1.

        """
        self.descriptor = descriptor
        self.selector = selector
        self.drop_table = drop_table
        self.class_key = class_key

        ratios_sum = get_ratio_list_sum([s.ratio for s in sets])
        set_names = [s.name for s in sets]
        if len(set_names) != len(set(set_names)):
            raise ValueError(f"Set names must be unique, got {set_names}")

        if ratios_sum != 1.0:
            sets.append(GoldSet(name="not assigned", ratio=1.0 - ratios_sum))

        self.sets = sets

        self.selector.collate_fn = pxt_torch_dataset_collate_fn

    def split(self, dataset: Dataset) -> dict[str, set[int]]:
        """Split the dataset into multiple sets based on the configured ratios.

        The dataset is first described using the gold descriptor (extracts features), and then samples are selected
        for each set based on the specified ratios. If a class_key is provided, stratified sampling is performed
        to maintain class distribution across sets.

        The last set and potentially the last class of each set in case of stratification will take the remaining samples
        to avoid rounding issues.

        Args:
            dataset: Dataset to be split.

        Returns:
            A dictionary mapping set names to sets of sample indices assigned to each set.

        Raises:
            ValueError: If any set results in zero samples due to its ratio, if class_key is not found,
            or if class stratification results in zero samples for any class in a set.
        """
        described_table = self.descriptor.describe(dataset)
        described_table.add_column(gold_set=pxt.String, if_exists="error")

        class_expr = self._get_class_expr(described_table)
        class_ratios = _get_column_distinct_ratios(described_table, class_expr)

        sample_count = described_table.count()

        # select data for all sets except the last one
        for idx_set, gold_set in enumerate(self.sets[:-1]):
            set_count = int(gold_set.ratio * sample_count)
            if set_count == 0:
                raise ValueError(
                    f"Set '{gold_set.name}' has ratio {gold_set.ratio} which results "
                    f"in zero samples for dataset of size {sample_count}."
                )
            self._select_for_set(
                described_table,
                class_ratios,
                gold_set.name,
                set_count,
                class_expr,
            )

        # remaining samples are assigned to the last set
        remaining_idx_list = [
            row["idx"]
            for row in described_table.select(described_table.idx)
            .where(described_table.gold_set == None)  # noqa: E711
            .collect()
        ]
        set_value_to_idx_rows(
            described_table,
            described_table.gold_set,
            set(remaining_idx_list),
            self.sets[-1].name,
        )

        self._drop_tables()

        return {
            gold_set.name: set(
                row["idx"]
                for row in described_table.where(
                    described_table.gold_set == gold_set.name
                )
                .select(described_table.idx)
                .collect()
            )
            for gold_set in self.sets
        }

    def _get_class_expr(self, described_table: pxt.Table) -> Expr:
        if (
            self.class_key is not None
            and self.class_key not in described_table.columns()
        ):
            raise ValueError(
                f"Class key '{self.class_key}' not found in described table columns: {described_table.columns()}"
            )

        class_key = self.class_key if self.class_key is not None else "gold_split_class"
        if self.class_key is None:
            described_table.add_column(**{class_key: pxt.String}, if_exists="error")

        return get_expr_from_column_name(described_table, class_key)

    def _select_for_set(
        self,
        described_table: Table,
        class_ratios: dict[str, float],
        set_name: str,
        set_count: int,
        class_expr: Expr,
    ) -> None:
        for class_idx, (class_label, class_ratio) in enumerate(class_ratios.items()):
            if class_idx < len(class_ratios) - 1:
                class_count = int(set_count * class_ratio)
                if class_count == 0:
                    raise ValueError(
                        f"Class '{class_label}' has ratio {class_ratio} which results "
                        f"in zero samples for set '{set_name}' with size {set_count}."
                    )
            else:
                # The last class takes the missing samples count to avoid rounding issues
                existing_count = described_table.where(
                    described_table.gold_set == set_name
                ).count()
                class_count = set_count - existing_count

            class_table = described_table.where(
                (class_expr == class_label) & (described_table.gold_set == None)  # noqa: E711
            ).select()

            torch_dataset = ResetableTorchIterableDataset(
                GoldPxtTorchDataset(
                    class_table, shapes=get_array_column_shapes(described_table)
                )
            )

            selected_indices = self.selector.select(torch_dataset, class_count)
            described_table.where(described_table.idx.isin(selected_indices)).update(
                {"gold_set": set_name}
            )

    def _drop_tables(self) -> None:
        if self.drop_table:
            for table_name in (self.descriptor.table_path, self.selector.table_path):
                pxt.drop_table(table_name)



from typing import Iterable


def check_x_and_y_shapes(x_shape: tuple[int, ...], y_shape: tuple[int, ...]) -> None:
    """Check compatibility of shapes of x and y tensors.

    Args:
        x_shape: Shape of the input tensor x.
        y_shape: Shape of the target tensor y.

    Raises:
        ValueError: If the shapes are incompatible. See the conditions below for details.
    Conditions:
        - If x is 1D, x and y must have the same shape.
        - If x is 2D or more, y must have a channel dimension of 1 (i.e., y_shape[1] == 1).
        - If y has a batch size greater than 1 (i.e., y_shape[0] > 1), x and y must have the same batch size (i.e., x_shape[0] == y_shape[0]).
        - If x has more than 2 dimensions, the additional (after channel) dimensions of x and y must match.
    """

    if len(x_shape) == 1:
        if x_shape != y_shape:
            raise ValueError("x and y must have the same shape when x is 1D")
    else:
        if y_shape[1] != 1:
            raise ValueError(
                "x and y must have the same channel dimension when x is 2D"
            )

        batch_y = y_shape[0]
        if batch_y > 1 and x_shape[0] != batch_y:
            raise ValueError(
                "x and y must have the same batch size when y batch size > 1"
            )

        if len(x_shape) > 2 and x_shape[2:] != y_shape[2:]:
            raise ValueError(
                "x and y must have compatible shapes when x is more than 2D"
            )


def get_ratio_list_sum(ratios: list[float]) -> float:
    """Get the sum of a list of ratios and validate it (total between 0 and 1).

    Args:
        ratios: A list of float ratios.

    Returns:
        The sum of the ratios.

    Raises:
        ValueError: If the sum of the ratios is not between 0 and 1 (included).
    """
    ratio_sum = sum(ratios)
    if not (0 < ratio_sum <= 1.0):
        raise ValueError("Sum of split ratios must be 1.0")

    return ratio_sum


def get_ratios_for_counts(counts: Iterable[int]) -> list[float]:
    """Get ratios for a list of counts.

    Args:
        counts: An iterable of integer counts.

    Returns:
        A list of float ratios corresponding to the input counts.
    """
    total = sum(counts)
    return [count / total for count in counts]



import torch
import pytest
from goldener.utils import check_x_and_y_shapes


class TestCheckXAndYShapes:
    def test_1d_shapes_match(self):
        x = torch.zeros(5)
        y = torch.zeros(5)
        check_x_and_y_shapes(x.shape, y.shape)  # Should not raise

    def test_1d_shapes_mismatch(self):
        x = torch.zeros(5)
        y = torch.zeros(6)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_2d_shapes_match_channel1(self):
        x = torch.zeros(3, 4)
        y = torch.zeros(3, 1)
        check_x_and_y_shapes(x.shape, y.shape)  # Should not raise

    def test_2d_shapes_mismatch_channel(self):
        x = torch.zeros(3, 4)
        y = torch.zeros(3, 2)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_2d_shapes_mismatch_batch(self):
        x = torch.zeros(3, 4)
        y = torch.zeros(4, 1)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_2d_shapes_match_batch_y_one(self):
        x = torch.zeros(3, 4)
        y = torch.zeros(1, 1)
        check_x_and_y_shapes(x.shape, y.shape)  # Should not raise

    def test_3d_shapes_match(self):
        x = torch.zeros(2, 3, 4)
        y = torch.zeros(2, 1, 4)
        check_x_and_y_shapes(x.shape, y.shape)  # Should not raise

    def test_3d_shapes_mismatch_channel(self):
        x = torch.zeros(2, 3, 4)
        y = torch.zeros(2, 2, 4)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_3d_shapes_mismatch_batch(self):
        x = torch.zeros(2, 3, 4)
        y = torch.zeros(3, 1, 4)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_3d_shapes_mismatch_last_dims(self):
        x = torch.zeros(2, 3, 4)
        y = torch.zeros(2, 1, 5)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_3d_shapes_match_batch_y_one(self):
        x = torch.zeros(2, 3, 4)
        y = torch.zeros(1, 1, 4)
        check_x_and_y_shapes(x.shape, y.shape)  # Should not raise

    def test_y_is_none(self):
        x = torch.zeros(2, 3, 4)
        y = None
        # Should not raise if y is None (skip check)
        if y is not None:
            check_x_and_y_shapes(x.shape, y.shape)



import pytest

import torch
from sklearn.decomposition import PCA

import pixeltable as pxt
from torch.utils.data import Dataset

from goldener.reduce import GoldReducer
from goldener.select import GoldSelector
from goldener.vectorize import GoldVectorizer


class DummyDataset(Dataset):
    def __init__(self, samples):
        self._samples = samples

    def __len__(self):
        return len(self._samples)

    def __getitem__(self, idx):
        return (
            self._samples[idx].copy()
            if isinstance(self._samples[idx], dict)
            else self._samples[idx]
        )


def collate_fn(batch):
    # simple collate that returns the single sample
    return batch[0]


class TestGoldSelector:
    def test_check_dataset_valid_no_collate(self):
        dataset = DummyDataset(
            [
                {"features": torch.tensor([1.0]), "idx": 0},
            ]
        )

        selector = GoldSelector(table_path="tmp.table", vectorizer=GoldVectorizer())
        # Should not raise
        selector._check_dataset(dataset)

    def test_check_dataset_valid_with_collate(self):
        dataset = DummyDataset(
            [
                {"features": torch.tensor([1.0]), "idx": 0},
            ]
        )

        selector = GoldSelector(
            table_path="tmp.table", vectorizer=GoldVectorizer(), collate_fn=collate_fn
        )
        # Should not raise
        selector._check_dataset(dataset)

    def test_check_dataset_non_dict(self):
        dataset = DummyDataset([1])
        selector = GoldSelector(table_path="tmp.table", vectorizer=GoldVectorizer())

        with pytest.raises(ValueError):
            selector._check_dataset(dataset)

    def test_check_dataset_missing_select_key(self):
        dataset = DummyDataset([{"not_features": torch.tensor([1.0]), "idx": 0}])
        selector = GoldSelector(table_path="tmp.table", vectorizer=GoldVectorizer())

        with pytest.raises(ValueError):
            selector._check_dataset(dataset)

    def test_check_dataset_select_not_tensor(self):
        dataset = DummyDataset([{"features": [1, 2, 3], "idx": 0}])
        selector = GoldSelector(table_path="tmp.table", vectorizer=GoldVectorizer())

        with pytest.raises(ValueError):
            selector._check_dataset(dataset)

    def test_check_dataset_missing_idx(self):
        dataset = DummyDataset([{"features": torch.tensor([1.0])}])
        selector = GoldSelector(table_path="tmp.table", vectorizer=GoldVectorizer())

        with pytest.raises(ValueError):
            selector._check_dataset(dataset)

    def test_initialize_table(self):
        table_path = "unit_test.test_select_initialize"

        sample = {"features": torch.rand(5, 2), "idx": 7}
        dataset = DummyDataset([sample])

        selector = GoldSelector(
            table_path=table_path,
            vectorizer=GoldVectorizer(),
            if_exists="replace_force",
        )

        pxt_table = selector._initialize_table(dataset)

        assert pxt_table.count() == 0
        assert set(pxt_table.columns()) == {
            "vector",
            "sample_idx",
            "idx",
            "selected",
            "chunked",
        }

        # Cleanup: drop the created table to avoid polluting environment
        try:
            pxt.drop_table(table_path)
        except Exception:
            # best effort cleanup
            pass

    def test_sequential_store_vectors_in_table(self):
        table_path = "unit_test.test_select_store_vectors"

        # two samples each with a single vector (1xD)
        sample0 = {
            "features": torch.rand(
                5,
                2,
            ),
            "idx": 0,
        }
        sample1 = {
            "features": torch.rand(
                5,
                2,
            ),
            "idx": 1,
        }
        dataset = DummyDataset([sample0, sample1])

        selector = GoldSelector(
            table_path=table_path,
            vectorizer=GoldVectorizer(),
            if_exists="replace_force",
        )

        pxt_table = selector._sequential_store_vectors_in_table(dataset)

        # Expect two vectors inserted (one per sample)
        assert pxt_table.count() == 4

        # Cleanup
        try:
            pxt.drop_table(table_path)
        except Exception:
            pass

    def test_sequential_store_vectors_in_table_with_collate_fn(self):
        table_path = "unit_test.test_select_store_vectors"

        def collate_fn(batch):
            return {
                "features": torch.stack([item["features"] for item in batch]),
                "idx": torch.tensor([item["idx"] for item in batch]),
            }

        # two samples each with a single vector (1xD)
        sample0 = {
            "features": torch.rand(
                5,
                2,
            ),
            "idx": 0,
        }
        sample1 = {
            "features": torch.rand(
                5,
                2,
            ),
            "idx": 1,
        }
        dataset = DummyDataset([sample0, sample1])

        selector = GoldSelector(
            table_path=table_path,
            vectorizer=GoldVectorizer(),
            collate_fn=collate_fn,
            if_exists="replace_force",
        )

        pxt_table = selector._sequential_store_vectors_in_table(dataset)

        # Expect two vectors inserted (one per sample)
        assert pxt_table.count() == 4

        # Cleanup
        try:
            pxt.drop_table(table_path)
        except Exception:
            pass

    def test_sequential_select(self):
        """Test _sequential_select using a real PixelTable table.

        This test will be skipped if the `pixeltable` package is not available.
        We only monkeypatch `_coresubset_selection` to control the selected indices
        while keeping the real table creation and insertion logic.
        """
        table_path = "unit_test.test_select_sequential"

        # prepare two samples, each with a single vector (shape: (1, D))
        dataset = DummyDataset(
            [{"features": torch.rand(5, 2), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path,
            vectorizer=GoldVectorizer(),
            if_exists="replace_force",
        )

        # Run selection (sequential path)
        selected = selector.select(dataset, select_count=10)

        # We expect one sample to be selected (the one corresponding to the first vector chosen by fake_coresubset)
        assert isinstance(selected, set)
        assert len(selected) == 10

        # Cleanup created table
        try:
            pxt.drop_table(table_path)
        except Exception:
            pass

    def test_sequential_select_with_chunk(self):
        """Test _sequential_select using a real PixelTable table.

        This test will be skipped if the `pixeltable` package is not available.
        We only monkeypatch `_coresubset_selection` to control the selected indices
        while keeping the real table creation and insertion logic.
        """
        table_path = "unit_test.test_select_sequential"

        # prepare two samples, each with a single vector (shape: (1, D))
        dataset = DummyDataset(
            [{"features": torch.rand(5, 2), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path,
            vectorizer=GoldVectorizer(),
            if_exists="replace_force",
            chunk=21,
        )

        # Run selection (sequential path)
        selected = selector.select(dataset, select_count=10)

        # We expect one sample to be selected (the one corresponding to the first vector chosen by fake_coresubset)
        assert isinstance(selected, set)
        assert len(selected) == 10

        # Cleanup created table
        try:
            pxt.drop_table(table_path)
        except Exception:
            pass

    def test_sequential_reducer(self):
        """Test _sequential_select using a real PixelTable table.

        This test will be skipped if the `pixeltable` package is not available.
        We only monkeypatch `_coresubset_selection` to control the selected indices
        while keeping the real table creation and insertion logic.
        """
        table_path = "unit_test.test_select_sequential"

        # prepare two samples, each with a single vector (shape: (1, D))
        dataset = DummyDataset(
            [{"features": torch.rand(5, 2), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path,
            vectorizer=GoldVectorizer(),
            reducer=GoldReducer(PCA(n_components=3)),
            if_exists="replace_force",
        )

        # Run selection (sequential path)
        selected = selector.select(dataset, select_count=10)

        # We expect one sample to be selected (the one corresponding to the first vector chosen by fake_coresubset)
        assert isinstance(selected, set)
        assert len(selected) == 10

        # Cleanup created table
        try:
            pxt.drop_table(table_path)
        except Exception:
            pass

    def test_select_with_target_key(self):
        """Ensure selection works when features to select are provided under a custom target key.

        The dataset samples include a `target` tensor which should be used by the selector
        when `select_target_key="target"` is passed.
        """
        table_path = "unit_test.test_select_with_target"

        # prepare samples that include both features and a target tensor
        dataset = DummyDataset(
            [
                {
                    "features": torch.rand(5, 2),
                    "target": torch.ones(
                        1,
                        2,
                    ),
                    "idx": idx,
                }
                for idx in range(50)
            ]
        )

        selector = GoldSelector(
            table_path=table_path,
            vectorizer=GoldVectorizer(),
            select_target_key="target",
            if_exists="replace_force",
        )

        selected = selector.select(dataset, select_count=10)

        assert isinstance(selected, set)
        assert len(selected) == 10

        # Cleanup created table
        try:
            pxt.drop_table(table_path)
        except Exception:
            pass



from dataclasses import dataclass
from functools import partial

from torch.utils.data import RandomSampler
from typing_extensions import assert_never
from typing import Callable

from enum import Enum

import torch
from torch import Generator

from goldener.torch_utils import make_2d_tensor
from goldener.utils import check_x_and_y_shapes


class FilterLocation(Enum):
    START = "start"
    END = "end"
    RANDOM = "random"


class Filter2DWithCount:
    """Filter 2D tensor rows based on specified criteria.

    Attributes:
        _filter_location: Location to filter from (start, end, random).
        _filter_count: Number of rows to filter.
        _keep: Whether to keep or remove the filtered rows.
        _random_sampler: Sampler for random filtering.
    """

    def __init__(
        self,
        filter_count: int = 1,
        filter_location: FilterLocation = FilterLocation.RANDOM,
        keep: bool = False,
        generator: Generator | None = None,
    ) -> None:
        """Initialize the Filter2DWithCount.

        Args:
            filter_count: Number of rows to filter.
            filter_location: Location to filter from (start, end, random).
            keep: Whether to keep or remove the filtered rows.
            generator: Random number generator for random filtering.
        """
        if filter_count <= 0:
            raise ValueError("filter_count must be greater than 0")

        self._filter_location = filter_location
        self._filter_count = filter_count
        self._keep = keep

        if self._filter_location == FilterLocation.RANDOM:
            self._random_sampler = partial(
                RandomSampler,
                replacement=False,
                num_samples=self._filter_count,
                generator=generator,
            )

    @property
    def is_random(self):
        """Check if the filter is random."""
        return self._filter_location is FilterLocation.RANDOM

    def filter(
        self,
        x: torch.Tensor,
    ) -> torch.Tensor:
        return self.filter_tensors({"x": x})["x"]

    def filter_tensors(
        self,
        x: dict[str, torch.Tensor],
    ) -> dict[str, torch.Tensor]:
        """Filter the input tensor or dictionary of tensors.

        Args:
            x: Input 2D tensor or dictionary of 2D tensors to filter. In case of a dictionary,
            all tensors must have the same number of rows and this number must be greater than filter_count.

        Returns:
            Filtered tensor or dictionary of tensors. If the batch size is less than filter_count,
            the input is returned unchanged.

        Raises:
            ValueError: If input tensors do not have the same batch size or are not 2 dimensional.
        """
        first_tensor = next(iter(x.values()))
        batch_size = first_tensor.shape[0]
        for tensor in x.values():
            if tensor.ndim != 2:
                raise ValueError(
                    "All input tensors must be 2D to filter them with Filter2DWithCount."
                )

            if tensor.shape[0] != batch_size:
                raise ValueError(
                    "All input tensors must have the same batch size to filter them with Filter2DWithCount."
                )

        if batch_size < self._filter_count:
            return x

        if self._filter_location is FilterLocation.START:
            return self._start_filter(x)

        elif self._filter_location is FilterLocation.END:
            return self._end_filter(x)

        elif self._filter_location is FilterLocation.RANDOM:
            return self._random_filter(x)

        else:
            assert_never(self._filter_location)

    def _start_filter(self, x: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        return (
            {k: v[: self._filter_count] for k, v in x.items()}
            if self._keep
            else {k: v[self._filter_count :] for k, v in x.items()}
        )

    def _end_filter(self, x: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        return (
            {k: v[-self._filter_count :] for k, v in x.items()}
            if self._keep
            else {k: v[: -self._filter_count] for k, v in x.items()}
        )

    def _random_filter(self, x: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        first_value = next(iter(x.values()))
        indices = list(self._random_sampler(range(len(first_value))))
        if self._keep:
            return {k: v[indices] for k, v in x.items()}
        else:
            mask = torch.ones(len(first_value), dtype=torch.bool)
            mask[indices] = 0
            return {k: v[mask.bool(), :] for k, v in x.items()}


@dataclass
class Vectorized:
    """Dataclass to hold vectorized tensors and the corresponding batch indices.

    Attributes:
        vectors: 2D tensor of vectorized data.
        batch_indices: 1D tensor containing information about the origin of each vector.
    """

    vectors: torch.Tensor
    batch_indices: torch.Tensor


class GoldVectorizer:
    """Transform input as 2D tensor and filter based on target tensor.

    Attributes:
        keep: Filter2DWithCount instance to keep specific rows in the input `x` of `filter`.
        remove: Filter2DWithCount instance to remove specific rows in the input `x` of `filter`.
        random_filter: Random Filter2DWithCount instance to randomly filter vectors randomly after
        applying `keep`, `remove` and the target `y` on the input `x` of filter.
        transform_y: Optional callable to transform the target tensor before transforming it to 2D.
    """

    def __init__(
        self,
        keep: Filter2DWithCount | None = None,
        remove: Filter2DWithCount | None = None,
        random_filter: Filter2DWithCount | None = None,
        transform_y: Callable[[torch.Tensor], torch.Tensor] | None = None,
    ) -> None:
        """Initialize the Vectorizer."""
        self.transform_y = transform_y

        if keep is not None and keep.is_random:
            raise ValueError("The 'keep' filter cannot be random.")
        self.keep = keep

        if remove is not None and remove.is_random:
            raise ValueError("The 'remove' filter cannot be random.")
        self.remove = remove

        if random_filter is not None and not random_filter.is_random:
            raise ValueError("The 'random_filter' must be random.")
        self.random_filter = random_filter

    def vectorize(
        self,
        x: torch.Tensor,
        y: torch.Tensor | None = None,
    ) -> Vectorized:
        """Vectorize input tensor and filter based on target tensor.

        If y is not provided, only filtering and vectorization are performed. If y is provided,
        it is used to filter the vectorized x tensor.

        Args:
            x: Input tensor to vectorize.
            y: Optional target tensor to filter the input tensor.

        Returns:
            Vectorized and filtered input tensor with the information
            about which element of the batch it corresponds to.

        Raises:
            ValueError: If x and y shapes are incompatible. See check_x_and_y_shapes for details.
        """
        if x.ndim < 3:
            raise ValueError("Input tensor x to vectorize must be at least 3D.")

        if y is not None and self.transform_y is not None:
            y = self.transform_y(y)

        filtered_x = []
        filtered_batch_info = []
        for idx_sample, x_sample in enumerate(x):
            x_sample = x_sample.unsqueeze(0)
            x_sample = make_2d_tensor(x_sample)

            x_sample = self._apply_filter(
                self.keep.filter if self.keep is not None else None, x_sample
            )
            x_sample = self._apply_filter(
                self.remove.filter if self.remove is not None else None, x_sample
            )

            if y is not None:
                y_sample = y[idx_sample].unsqueeze(0)
                x_sample = self._filter_2d_tensors_from_y(x_sample, y_sample)

            filtered_x.append(x_sample)
            filtered_batch_info.append(
                torch.full_like(x_sample[:, 0], idx_sample, dtype=torch.long)
            )

        return Vectorized(
            torch.cat(filtered_x, dim=0), torch.cat(filtered_batch_info, dim=0)
        )

    def _apply_filter(
        self,
        filter: Callable[[torch.Tensor], torch.Tensor] | None,
        x: torch.Tensor,
    ) -> torch.Tensor:
        if filter is None:
            return x

        return filter(x)

    def _filter_2d_tensors_from_y(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        """Filter 2D tensor rows based on target tensor.

        When y is not the same batch size as x, it is expanded to match x's batch size.

        Args:
            x: Input 2D tensor to filter. It can be the output of `make_2d_tensor`.
            y: Target tensor to filter the input tensor. It is still in its raw configuration.
            The transform_y callable is applied to y before transforming it to a 2D tensor.

        Returns:
            Filtered input tensor.

        Raises:
            ValueError: If x and y shapes (after 2D transformation) are incompatible.
            See check_x_and_y_shapes for details.
            ValueError: If y tensor after transform contains only zeros.
        """
        y = make_2d_tensor(y)

        y_shape = y.shape
        x_shape = x.shape
        check_x_and_y_shapes(x_shape, y_shape)

        if (y == 0).all():
            raise ValueError(
                "The y tensor after transform must contain at least one "
                "non-zero value to select vectors from x."
            )

        return x[y.bool().squeeze(-1)]



from collections import defaultdict
from typing import Literal, Any, Iterator, Sequence, Callable

import numpy as np
import pixeltable as pxt
from pixeltable import DataFrame
from pixeltable.catalog import Table
from pixeltable.exprs import Expr

import torch
from pixeltable.utils.pytorch import PixeltablePytorchDataset

from goldener.utils import get_ratios_for_counts


def get_expr_from_column_name(table: Table, column_name: str) -> Expr:
    """Get the expression object for a given column name in a PixelTable table.

    Args:
        table: The PixelTable table.
        column_name: The name of the column.

    Returns:
        pxt.Expr: The expression object corresponding to the column name.

    Raises:
        ValueError: If the column name does not exist in the table schema.
    """
    if column_name not in table.columns():
        raise ValueError(f"Column '{column_name}' does not exist in the table schema.")

    return getattr(table, column_name)


def create_pxt_dirs_for_path(table_path: str) -> None:
    """Create necessary PixelTable directories for a given table path.

    Args:
        table_path: The full path of the PixelTable table (e.g., 'dir1.dir2.table_name').
    """
    pxt_path_split = table_path.split(".")
    for pxt_dir_idx in range(len(pxt_path_split) - 1):
        pxt_dir = ".".join(pxt_path_split[0 : pxt_dir_idx + 1])
        pxt.create_dir(pxt_dir, if_exists="ignore")


def create_pxt_table_from_sample(
    table_path: str,
    sample: dict,
    if_exists: Literal["error", "replace_force"] = "error",
) -> Table:
    """Create a PixelTable table from a sample dictionary.

    Args:
        table_path: The full path of the PixelTable table to create (e.g., 'dir1.dir2.table_name').
        sample: A dictionary representing a single sample, where keys are column names and values are the corresponding data.
        if_exists: Behavior if the table already exists. Options are 'error' or 'replace_force'.
    """
    create_pxt_dirs_for_path(table_path)

    pxt_table = pxt.create_table(
        table_path,
        source=[
            {
                key: (value.numpy() if isinstance(value, torch.Tensor) else value)
                for key, value in sample.items()
            }
        ],
        if_exists=if_exists,
    )

    return pxt_table


def set_value_to_idx_rows(
    table: Table,
    col_expr: Expr,
    idx_list: set[int],
    value: int | float | str,
) -> None:
    """Set a column to a specific value for rows with given indices in a PixelTable table.

    Args:
        table: The PixelTable table. Must contain an 'idx' column.
        col_expr: The column expression to be set.
        idx_list: List of row indices to update.
        value: The value to set the column to.
    """
    table.where(table.idx.isin(idx_list)).update({col_expr.display_str(): value})


def update_column_if_too_many(
    table: Table,
    col_expr: Expr,
    value: int | float | str,
    max_count: int,
    new_value: Any,
) -> None:
    """Update some values of a column if the count of that value exceeds a maximum count.

    Args:
        table: The PixelTable table. Must contain an 'idx' column.
        col_expr: The column expression to be checked and updated.
        value: The value to check the count of.
        max_count: The maximum allowed count for the specified value.
        new_value: The new value to set for excess rows.
    """
    value_df = table.where(col_expr == value)
    value_count = value_df.select().count()
    if value_count > max_count:
        to_move = value_count - max_count
        indices = [row["idx"] for row in value_df.sample(to_move).collect()]
        set_value_to_idx_rows(table, col_expr, set(indices), new_value)


def pxt_torch_dataset_collate_fn(batch: list[dict[str, Any]]) -> dict[str, Any]:
    """Collate function for torch datasets obtained from a Pixeltable table.

    Args:
        batch: A list of samples, where each sample is a dictionary.

    Returns:
        A single dictionary with collated values. The numpy arrays and integers
        are stacked into torch tensors. All other types are kept as lists.
    """
    value_list_dict = defaultdict(list)
    conversion_dict: dict[str, Callable[[Sequence[Any]], torch.Tensor]] = {}

    def stack_arrays_and_convert_to_torch(arrays: Sequence[np.ndarray]) -> torch.Tensor:
        return torch.from_numpy(np.stack(arrays, axis=0))

    def stack_int_as_torch(ints: Sequence[int]) -> torch.Tensor:
        return torch.tensor(ints, dtype=torch.int64)

    for idx_sample, sample in enumerate(batch):
        for key, value in sample.items():
            value_list_dict[key].append(value)
            if idx_sample == 0:
                if isinstance(value, np.ndarray):
                    conversion_dict[key] = stack_arrays_and_convert_to_torch
                elif isinstance(value, int):
                    conversion_dict[key] = stack_int_as_torch

    return {
        key: conversion_dict[key](value) if key in conversion_dict else value
        for key, value in value_list_dict.items()
    }


class GoldPxtTorchDataset(PixeltablePytorchDataset):
    """A Pixeltable PyTorch Dataset that reshapes array columns to their original shapes.

    This class intends to solve a current issue in Pixeltable dataset where array columns
    are flattened when converted to PyTorch datasets. This class reshapes those columns
    back to their original shapes during iteration.

    Attributes:
        shapes: A dictionary mapping column names of arrays to their original shapes.
    """

    def __init__(
        self, source: Table | DataFrame, shapes: dict[str, tuple[int, ...]]
    ) -> None:
        """Initialize the GoldPxtTorchDataset.

        Args:
            source: The PixelTable Table or DataFrame to convert to a PyTorch dataset.
            shapes: A dictionary mapping column names of arrays to their original shapes.
        """
        torch_dataset = source.to_pytorch_dataset()
        assert isinstance(torch_dataset, PixeltablePytorchDataset)
        super().__init__(torch_dataset.path, torch_dataset.image_format)
        self.shapes = shapes

    def __iter__(self) -> Iterator[dict[str, Any]]:
        super_iterator = super().__iter__()
        for item in super_iterator:
            yield {
                key: value
                if key not in self.shapes
                else value.reshape(self.shapes[key])
                for key, value in item.items()
            }


def get_array_column_shapes(table: Table) -> dict[str, tuple[int, ...]]:
    """Get the shapes of array columns in a PixelTable table.

    Args:
        table: The PixelTable table.

    Returns:
        A dictionary mapping column names of array columns to their shapes.
    """
    shapes = {}
    for col_name in table.columns():
        col_expr = get_expr_from_column_name(table, col_name)
        col_expr_dict = col_expr.col_type.as_dict()
        if col_expr_dict["_classname"] == "ArrayType":
            shapes[col_name] = tuple(col_expr_dict["shape"])

    return shapes


def get_distinct_value_and_count_in_column(
    table: Table,
    col_expr: Expr,
) -> dict[Any, int]:
    """Get distinct values and their counts in a specified column of a PixelTable table.

    Args:
        table: The PixelTable table.
        col_expr: The column expression to analyze.

    Returns:
        A dictionary mapping distinct values to their counts in the specified column.
    """
    return {
        distinct_item[col_expr.display_str()]: table.where(
            col_expr == distinct_item[col_expr.display_str()]
        ).count()
        for distinct_item in table.select(col_expr).distinct().collect()
    }


def _get_column_distinct_ratios(table: Table, class_expr: Expr) -> dict[str, float]:
    """Get the ratios of distinct values in a specified column of a PixelTable table.

    Args:
        table: The PixelTable table.
        class_expr: The column expression to analyze.

    Returns:
        A dictionary mapping distinct values to their ratios in the specified column.
    """
    value_and_count = get_distinct_value_and_count_in_column(table, class_expr)
    ratios = get_ratios_for_counts(value_and_count.values())

    return {
        class_label: class_ratio
        for class_label, class_ratio in zip(value_and_count.keys(), ratios)
    }



import pytest
import torch

import pixeltable as pxt

from goldener.describe import GoldDescriptor
from goldener.extract import TorchGoldFeatureExtractorConfig, TorchGoldFeatureExtractor


class DummyModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(3, 4, 3, padding=1)
        self.relu = torch.nn.ReLU()

    def forward(self, x):
        x = self.conv1(x)
        x = self.relu(x)
        return x


@pytest.fixture
def extractor():
    model = DummyModel()
    config = TorchGoldFeatureExtractorConfig(model=model, layers=None)
    return TorchGoldFeatureExtractor(config)


class DummyDataset:
    def __init__(self, output_shape=(3, 2, 2)):
        # produce a fixed tensor
        self.output_shape = output_shape

    def __len__(self):
        return 2

    def __getitem__(self, idx):
        return {"data": torch.zeros(3, 8, 8), "idx": idx, "label": "dummy"}


class TestGoldDescriptor:
    def test_simple_usage(self, extractor):
        desc = GoldDescriptor(
            table_path="unit_test.test_describe",
            extractor=extractor,
            batch_size=2,
            collate_fn=None,
            device=torch.device("cpu"),
            if_exists="replace_force",
        )

        table = desc.describe(DummyDataset())

        assert table.count() == 2
        for i, row in enumerate(table.collect()):
            assert row["idx"] == i
            assert (row["data"] == torch.zeros(3, 8, 8).numpy()).all()
            assert row["label"] == "dummy"
            assert row["features"].shape == (4, 8, 8)

        try:
            pxt.drop_table(desc.table_path)
        except Exception:
            pass

    def test_without_idx(self, extractor):
        def collate_fn(batch):
            data = torch.stack([b["data"] for b in batch], dim=0)
            return {"data": data}

        desc = GoldDescriptor(
            table_path="unit_test.test_describe",
            batch_size=2,
            extractor=extractor,
            collate_fn=collate_fn,
            device=torch.device("cpu"),
            if_exists="replace_force",
        )

        table = desc.describe(
            DummyDataset(),
        )
        assert table.count() == 2
        for i, row in enumerate(table.collect()):
            assert row["idx"] == i

        try:
            pxt.drop_table(desc.table_path)
        except Exception:
            pass

    def test_describe_with_non_dict_item(self, extractor):
        # Dataset returning a non-dict should raise
        desc = GoldDescriptor(
            table_path="unit_test.test_describe",
            extractor=extractor,
            collate_fn=lambda x: [d["data"] for d in x],
            device=torch.device("cpu"),
            if_exists="replace_force",
        )
        with pytest.raises(ValueError):
            desc.describe(DummyDataset())

    def test_describe_missing_data_key(self, extractor):
        desc = GoldDescriptor(
            table_path="unit_test.test_describe",
            extractor=extractor,
            collate_fn=lambda x: {"not data": "not_data"},
            device=torch.device("cpu"),
            if_exists="replace_force",
        )
        with pytest.raises(ValueError):
            desc.describe(
                DummyDataset(),
            )

    def test_with_collate_fn(self, extractor):
        def collate_fn(batch):
            data = torch.stack([b["data"] for b in batch], dim=0)
            idxs = [b["idx"] for b in batch]
            labels = [b["label"] for b in batch]
            return {"data": data, "idx": idxs, "label": labels}

        desc = GoldDescriptor(
            table_path="unit_test.test_describe",
            batch_size=2,
            extractor=extractor,
            collate_fn=collate_fn,
            device=torch.device("cpu"),
            if_exists="replace_force",
        )

        table = desc.describe(
            DummyDataset(),
        )
        assert table.count() == 2
        for i, row in enumerate(table.collect()):
            assert row["idx"] == i
            assert row["label"] == "dummy"

        desc_table = pxt.get_table(desc.table_path)
        column_schema = desc_table.get_metadata()["columns"]
        for col_name, col_dict in column_schema.items():
            if col_name == "features":
                assert col_dict["type_"] == "Array[(4, 8, 8), Float]"
            elif col_name == "data":
                assert col_dict["type_"] == "Array[(3, 8, 8), Float]"
            elif col_name == "idx":
                assert col_dict["type_"] == "Int"
            elif col_name == "label":
                assert col_dict["type_"] == "String"

        try:
            pxt.drop_table(desc.table_path)
        except Exception:
            pass



import pytest
import torch

import pixeltable as pxt
from pixeltable import Error
from torch.utils.data import Dataset

from goldener.describe import GoldDescriptor
from goldener.extract import TorchGoldFeatureExtractorConfig, TorchGoldFeatureExtractor
from goldener.split import GoldSplitter, GoldSet
from goldener.vectorize import GoldVectorizer
from goldener.select import GoldSelector


class DummyModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(3, 4, 3, padding=1)
        self.relu = torch.nn.ReLU()

    def forward(self, x):
        x = self.conv1(x)
        x = self.relu(x)
        return x


@pytest.fixture
def extractor():
    model = DummyModel()
    config = TorchGoldFeatureExtractorConfig(model=model, layers=None)
    return TorchGoldFeatureExtractor(config)


class DummyDataset(Dataset):
    def __init__(self, samples):
        self._samples = samples

    def __len__(self):
        return len(self._samples)

    def __getitem__(self, idx):
        return (
            self._samples[idx].copy()
            if isinstance(self._samples[idx], dict)
            else self._samples[idx]
        )


@pytest.fixture
def descriptor(extractor):
    return GoldDescriptor(
        table_path="unit_test.test_split",
        extractor=extractor,
        batch_size=2,
        collate_fn=None,
        device=torch.device("cpu"),
        if_exists="replace_force",
    )


@pytest.fixture
def selector():
    return GoldSelector(
        table_path="unit_test.selector_split",
        vectorizer=GoldVectorizer(),
        if_exists="replace_force",
    )


@pytest.fixture
def basic_splitter(descriptor, selector):
    sets = [GoldSet(name="train", ratio=0.5), GoldSet(name="val", ratio=0.5)]
    return GoldSplitter(sets=sets, descriptor=descriptor, selector=selector)


class TestGoldSplitter:
    def test_basic_split(self, basic_splitter):
        splitted = basic_splitter.split(
            dataset=DummyDataset(
                [{"data": torch.rand(3, 8, 8), "idx": idx} for idx in range(10)]
            )
        )

        assert len(splitted) == 2
        assert len(splitted["train"]) == 5
        assert len(splitted["val"]) == 5

        for path in (
            basic_splitter.descriptor.table_path,
            basic_splitter.selector.table_path,
        ):
            try:
                pxt.drop_table(path)
            except Exception:
                pass

    def test_split_with_label(self, descriptor, selector):
        sets = [
            GoldSet(name="train", ratio=0.5),
        ]
        splitter = GoldSplitter(
            sets=sets, descriptor=descriptor, selector=selector, class_key="label"
        )

        splitted = splitter.split(
            dataset=DummyDataset(
                [
                    {"data": torch.rand(3, 8, 8), "idx": idx, "label": str(idx % 2)}
                    for idx in range(10)
                ]
            )
        )

        assert set(splitted.keys()) == {"train", "not assigned"}
        assert len(splitted["train"]) == 5
        assert len(splitted["not assigned"]) == 5

        for path in (descriptor.table_path, selector.table_path):
            try:
                pxt.drop_table(path)
            except Exception:
                pass

    def test_duplicated_set_names(self, descriptor, selector):
        sets = [GoldSet(name="train", ratio=0.5), GoldSet(name="train", ratio=0.3)]
        with pytest.raises(ValueError):
            GoldSplitter(sets=sets, descriptor=descriptor, selector=selector)

    def test_class_key_not_found(self, descriptor, selector):
        sets = [GoldSet(name="only", ratio=0.5)]
        splitter = GoldSplitter(
            sets=sets, descriptor=descriptor, selector=selector, class_key="nonexistent"
        )

        with pytest.raises(ValueError):
            splitter.split(DummyDataset([{"data": torch.rand(3, 8, 8), "idx": 0}]))

        for path in (descriptor.table_path, selector.table_path):
            try:
                pxt.drop_table(path)
            except Exception:
                pass

    def test_gold_split_class_existing(self, descriptor, selector):
        sets = [GoldSet(name="only", ratio=0.5)]
        splitter = GoldSplitter(
            sets=sets,
            descriptor=descriptor,
            selector=selector,
        )

        with pytest.raises(Error):
            splitter.split(
                DummyDataset(
                    [
                        {
                            "data": torch.rand(3, 8, 8),
                            "idx": 0,
                            "gold_split_class": "useless",
                        }
                    ]
                )
            )

        for path in (descriptor.table_path, selector.table_path):
            try:
                pxt.drop_table(path)
            except Exception:
                pass

    def test_set_with_0_population(self, descriptor, selector):
        sets = [GoldSet(name="only", ratio=0.1)]
        splitter = GoldSplitter(sets=sets, descriptor=descriptor, selector=selector)

        with pytest.raises(ValueError):
            splitter.split(
                dataset=DummyDataset(
                    [{"data": torch.rand(3, 8, 8), "idx": idx} for idx in range(1)]
                )
            )

        for path in (descriptor.table_path, selector.table_path):
            try:
                pxt.drop_table(path)
            except Exception:
                pass

    def test_class_with_0_population(self, descriptor, selector):
        sets = [GoldSet(name="only", ratio=0.1)]
        splitter = GoldSplitter(
            sets=sets, descriptor=descriptor, selector=selector, class_key="label"
        )

        with pytest.raises(ValueError):
            splitter.split(
                dataset=DummyDataset(
                    [
                        {
                            "data": torch.rand(3, 8, 8),
                            "idx": idx,
                            "label": "A" if idx < 5 else "B",
                        }
                        for idx in range(10)
                    ]
                )
            )

        for path in (descriptor.table_path, selector.table_path):
            try:
                pxt.drop_table(path)
            except Exception:
                pass



from typing import Callable, Literal

import pixeltable as pxt
from pixeltable.catalog import Table

import torch
import jax.numpy as jnp

from coreax import SquaredExponentialKernel, Data
from coreax.kernels import median_heuristic
from coreax.solvers import KernelHerding
from torch.utils.data import Dataset

from goldener.pxt_utils import create_pxt_table_from_sample, set_value_to_idx_rows
from goldener.reduce import GoldReducer
from goldener.torch_utils import ResetableTorchIterableDataset
from goldener.vectorize import GoldVectorizer


class GoldSelector:
    """Selector class to select a subset of data points from a dataset using coresubset selection.

    If the dataset is too big to fit into memory or the coresubset selection algorithm is too
    computationally expensive, the coresubset selection can be performed in chunks.
    During the whole selection process, the vectors extracted from the dataset are stored
    in a PixelTable table. This table will contain the following columns:
        - idx: Index of the vector in the table.
        - vector: The vector representation of the data point.
        - sample_idx: Index of the original data point in the dataset.
        - selected: Boolean indicating whether the vector has been selected.
        - chunked: Boolean indicating whether the vector has been processed in a chunk.

    Attributes:
        table_path: Path to store the PixelTable table. It is used to store the vectors
        extracted from the dataset.
        vectorizer: Vectorizer to convert data points into vectors.
        select_key: Key in the dataset sample dictionary to select data points from.
        select_target_key: Optional key in the dataset sample dictionary to select target values from.
        It allows to filter out some vectors based on their target values.
        drop_table: Whether to drop the PixelTable table after selection.
        reducer: Optional dimensionality reducer to apply before selection.
        chunk: Optional chunk size for processing data in chunks.
        collate_fn: Optional collate function for the DataLoader.
        batch_size: Batch size for the DataLoader.
        num_workers: Number of workers for the DataLoader.
        if_exists: Behavior if the PixelTable table already exists. Options are 'error' or 'replace_force'.
        distribute: Whether to use distributed selection.
        shuffle: Whether to shuffle the dataset during loading.
        generator: Optional random number generator for shuffling.
    """

    def __init__(
        self,
        table_path: str,
        vectorizer: GoldVectorizer,
        select_key: str = "features",
        select_target_key: str | None = None,
        drop_table: bool = False,
        reducer: GoldReducer | None = None,
        chunk: int | None = None,
        collate_fn: Callable | None = None,
        batch_size: int | None = None,
        num_workers: int | None = None,
        if_exists: Literal["error", "replace_force"] = "error",
        distribute: bool = False,
        shuffle: bool = False,
        generator: torch.Generator | None = None,
    ) -> None:
        self.table_path = table_path
        self.vectorizer = vectorizer
        self.select_key = select_key
        self.select_target_key = select_target_key
        self.drop_table = drop_table
        self.reducer = reducer
        self.chunk = chunk
        self.collate_fn = collate_fn
        self.if_exists = if_exists
        self.distribute = distribute
        self.shuffle = shuffle
        self.generator = generator

        self.batch_size: int | None
        self.num_workers: int | None

        if not self.distribute:
            self.batch_size = 1 if batch_size is None else batch_size
            self.num_workers = 0 if num_workers is None else num_workers
        else:
            self.batch_size = batch_size
            self.num_workers = num_workers

    def select(self, dataset: Dataset, select_count: int) -> set[int]:
        """Select a subset of data points from the dataset.

        The selection is done from a coresubset selection algorithm applied on the vectorized
        representation of the data points (filtered with the target if provided). When the chunk
        attribute is set, the selection is performed in chunks to reduce memory consumption. As well,
        if a reducer is provided, the vectors are reduced in dimension before applying the coresubset selection.

        All along the selection process, the vectors are stored in a PixelTable table located at table_path.
        If drop_table is set to True, the table is deleted after the selection.

        Args:
            dataset: Dataset to select from. Each item should be a dictionary with
            at least the `select_key` and `idx` keys after applying the collate_fn.
            If the collate_fn is None, the dataset is expected to directly provide such batches.
            select_count: Number of data points to select.
        """
        self._check_dataset(dataset)

        if self.distribute:
            selected = self._distributed_select(dataset, select_count)
        else:
            selected = self._sequential_select(dataset, select_count)

        if self.drop_table:
            pxt.drop_table(self.table_path)

        return selected

    def _get_selected_indices(self, table: Table) -> set[int]:
        return set(
            [
                row["sample_idx"]
                for row in table.where(table.selected == True)  # noqa: E712
                .select(table.sample_idx)
                .distinct()
                .collect()
            ]
        )

    def _sequential_select(self, dataset: Dataset, select_count: int) -> set[int]:
        vector_table = self._sequential_store_vectors_in_table(dataset)

        # The coresubset selection is done from all the patches (after filtering) of all data point.
        # Then, the same data point can be selected multiple times if it has multiple patches selected.
        # To achieve select_count of unique data points, we loop until we have enough unique data points selected.
        already_selected = 0
        while already_selected < select_count:
            # select only rows still not selected
            to_chunk_from = vector_table.where(vector_table.selected == False)  # noqa: E712
            to_chunk_from.update(
                {"chunked": False}
            )  # unchunk all rows not yet selected

            # initialize the chunk settings: chunk size, number of chunks, selection per chunk
            to_chunk_from_count = to_chunk_from.count()
            chunk_size = (
                to_chunk_from_count
                if self.chunk is None
                else min(self.chunk, to_chunk_from_count)
            )
            chunk_loop_count = to_chunk_from_count // chunk_size
            select_per_chunk = (select_count - already_selected) // chunk_loop_count
            if select_per_chunk == 0:
                select_per_chunk = 1

            # make coresubset selection per chunk
            for chunk_idx in range(chunk_loop_count):
                if already_selected >= select_count:
                    break

                # select data for the current chunk among vector not yet selected
                to_select_from = vector_table.where(
                    (vector_table.chunked == False) & (vector_table.selected == False)  # noqa: E712
                ).select(vector_table.vector, vector_table.idx)
                if chunk_idx < chunk_loop_count - 1:
                    to_select_from = to_select_from.sample(chunk_size)

                # load the vectors and the corresponding indices for the chunk
                to_select = [
                    (
                        torch.from_numpy(sample["vector"]),
                        torch.tensor(sample["idx"]).unsqueeze(0),
                    )
                    for sample in to_select_from.collect()
                ]
                vectors_list, indices_list = map(list, zip(*to_select))
                vectors = torch.stack(vectors_list, dim=0)
                indices = torch.cat(indices_list, dim=0)
                set_value_to_idx_rows(
                    vector_table,
                    vector_table.chunked,
                    set(indices.tolist()),
                    True,
                )  # selected indices are marked as already chunked

                # make coresubset selection for the chunk
                if self.reducer is not None:
                    vectors = self.reducer.fit_transform(vectors)

                coresubset_indices = self._coresubset_selection(
                    vectors, select_per_chunk, indices
                )

                # update table with selected indices
                set_value_to_idx_rows(
                    vector_table,
                    vector_table.selected,
                    coresubset_indices,
                    True,
                )

                already_selected = len(self._get_selected_indices(vector_table))

        return self._get_selected_indices(vector_table)

    def _distributed_select(self, dataset: Dataset, select_count: int):
        raise NotImplementedError("Distributed selection is not implemented yet.")

    def _check_dataset(self, dataset: Dataset) -> None:
        sample = (
            next(dataset)
            if isinstance(dataset, ResetableTorchIterableDataset)
            else dataset[0]
        )
        if isinstance(dataset, ResetableTorchIterableDataset):
            dataset.reset()

        if self.collate_fn is not None:
            sample = self.collate_fn([sample])

        if not isinstance(sample, dict):
            raise ValueError(
                "Dataset sample must be a dictionary after applying the collate_fn."
            )

        if self.select_key not in sample:
            raise ValueError(f"Dataset sample must contain a {self.select_key} key.")
        elif not isinstance(sample[self.select_key], torch.Tensor):
            raise ValueError(
                f"Value of {self.select_key} must correspond to a torch.Tensor."
            )

        if "idx" not in sample:
            raise ValueError("Dataset must have an 'idx' key.")

        if self.select_target_key is not None:
            if self.select_target_key not in sample:
                raise ValueError(
                    f"Dataset sample must contain a {self.select_target_key} key."
                )
            elif not isinstance(sample[self.select_target_key], torch.Tensor):
                raise ValueError(
                    f"Value of {self.select_target_key} must correspond to a torch.Tensor."
                )

    def _initialize_table(
        self, dataset: Dataset | ResetableTorchIterableDataset
    ) -> Table:
        sample = (
            next(dataset)
            if isinstance(dataset, ResetableTorchIterableDataset)
            else dataset[0]
        )
        if isinstance(dataset, ResetableTorchIterableDataset):
            dataset.reset()

        if self.collate_fn is not None:
            sample = self.collate_fn([sample])
        elif sample[self.select_key].ndim == 2:
            sample[self.select_key] = sample[self.select_key].unsqueeze(0)

        vectorized = self.vectorizer.vectorize(
            sample[self.select_key],
            None
            if self.select_target_key is None
            else sample[self.select_target_key].unsqueeze(0),
        )

        unique_row = {
            "vector": vectorized.vectors[0].numpy(),
            "sample_idx": vectorized.batch_indices[0].item(),
            "idx": 0,
            "selected": False,
            "chunked": False,
        }

        pxt_table = create_pxt_table_from_sample(
            self.table_path, unique_row, self.if_exists
        )
        pxt_table.where(pxt_table.idx == 0).delete()  # remove the initial sample

        return pxt_table

    def _sequential_store_vectors_in_table(self, dataset: Dataset) -> Table:
        pxt_table = self._initialize_table(dataset)

        assert self.batch_size is not None
        assert self.num_workers is not None
        data_loader = torch.utils.data.DataLoader(
            dataset,
            batch_size=self.batch_size,
            collate_fn=self.collate_fn,
            num_workers=self.num_workers,
            shuffle=self.shuffle,
            generator=self.generator,
        )

        vector_count = 0
        for batch_idx, batch in enumerate(data_loader):
            vectors = batch[self.select_key]

            vectorized = self.vectorizer.vectorize(vectors)
            to_insert = [
                {
                    "idx": vector_idx + vector_count,
                    "vector": vector.numpy(),
                    "sample_idx": batch["idx"][sample_idx].item(),
                    "selected": False,
                    "chunked": False,
                }
                for vector_idx, (sample_idx, vector) in enumerate(
                    zip(vectorized.batch_indices, vectorized.vectors)
                )
            ]
            pxt_table.insert(to_insert)
            vector_count += len(to_insert)

        return pxt_table

    def _coresubset_selection(
        self, x: torch.Tensor, select_count: int, indices: torch.Tensor
    ) -> set[int]:
        herding_solver = KernelHerding(
            select_count,
            kernel=SquaredExponentialKernel(
                length_scale=float(median_heuristic(jnp.asarray(x.mean(1).numpy())))
            ),
        )
        herding_coreset, _ = herding_solver.reduce(Data(jnp.array(x.numpy())))  # type: ignore[arg-type]

        return set(
            indices[torch.tensor(herding_coreset.unweighted_indices.tolist())].tolist()
        )



from typing import Callable, Literal

import torch

from pixeltable.catalog import Table
from torch.utils.data import Dataset

from goldener.extract import GoldFeatureExtractor
from goldener.pxt_utils import create_pxt_table_from_sample


class GoldDescriptor:
    """Describe the `data` of a dataset by extracting features and storing them in a PixelTable table.

    Assuming all the data will not fit in memory, the dataset is processed in batches.
    All the data of the dataset, the computed features included, will be saved in a local Pixeltable table.
    The `data` of the dataset will be saved in the shape and scale obtained after applying the `collate_fn` if provided.
    These arrays are expected to be all the same size. All torch tensors will be converted to numpy arrays before saving.

    Attributes:
        table_path: Path to the PixelTable table where the description will be saved locally.
        extractor: FeatureExtractor instance for feature extraction.
        collate_fn: Optional function to collate dataset samples into batches composed of
        dictionaries with at least the `data` key returning a pytorch Tensor.
        If None, the dataset is expected to directly provide such batches. It should as well format the `data` value
        in the format expected by the feature extractor.
        batch_size: Optional batch size for processing the dataset.
        num_workers: Optional number of worker threads for data loading.
        if_exists: Behavior if the table already exists ('error' or 'replace_force'). If 'replace_force',
        the existing table will be replaced, otherwise an error will be raised.
        distribute: Whether to use distributed processing for feature extraction and table population. Not implemented yet.
    """

    def __init__(
        self,
        table_path: str,
        extractor: GoldFeatureExtractor,
        collate_fn: Callable | None = None,
        batch_size: int | None = None,
        num_workers: int | None = None,
        if_exists: Literal["error", "replace_force"] = "error",
        distribute: bool = False,
        device: torch.device | None = None,
    ):
        self.table_path = table_path
        self.extractor = extractor
        self.collate_fn = collate_fn
        self.if_exists = if_exists
        self.distribute = distribute

        if device is None:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.device = device

        self.batch_size: int | None
        self.num_workers: int | None
        if not self.distribute:
            self.batch_size = 1 if batch_size is None else batch_size
            self.num_workers = 0 if num_workers is None else num_workers
        else:
            self.batch_size = batch_size
            self.num_workers = num_workers

    def describe(self, dataset: Dataset) -> Table:
        """Describe the dataset by extracting features and storing them in a PixelTable table.

        Args:
            dataset: Dataset to be described. Each item should be a dictionary with at least the `data` key
            after applying the collate_fn. If the collate_fn is None, the dataset is expected to directly
            provide such batches.
        """
        sample = dataset[0]
        if self.collate_fn is not None:
            sample = self.collate_fn([sample])

        if not isinstance(sample, dict):
            raise ValueError(
                "Dataset items must be dictionaries after applying the collate_fn."
            )

        if "data" not in sample:
            raise ValueError("Dataset items must contain a 'data' key.")

        pxt_table = self._initialize_table(sample)

        if self.distribute:
            self._distributed_describe(pxt_table, dataset)
        else:
            pxt_table = self._sequential_describe(pxt_table, dataset)

        return pxt_table

    def _initialize_table(self, sample: dict) -> Table:
        """Initialize the PixelTable table using a single sample from the dataset.

        It creates the required PixelTable directories and table schema based on the provided sample.

        Args:
            sample: A single sample from the dataset to initialize the table schema.
            It should be a dictionary with at least the `data` key. The  `data` value must be formatted
            in the format expected by the feature extractor.
        """
        sample["features"] = self.extractor.extract_and_fuse(
            sample["data"].to(device=self.device)
        )
        if self.collate_fn is not None:
            for key, value in sample.items():
                if isinstance(value, torch.Tensor):
                    # Only initial tensors will have a batch dimension added by the collate_fn
                    # otherwise, they were initially single values (float or int)
                    if value.ndim > 1:
                        sample[key] = value.squeeze(0)
                    else:
                        sample[key] = value.item()
                else:
                    # non tensor values are expected to be lists of single values
                    sample[key] = value[0]

        if "idx" not in sample:
            sample["idx"] = 0

        pxt_table = create_pxt_table_from_sample(
            self.table_path, sample, self.if_exists
        )
        pxt_table.where(pxt_table.idx == 0).delete()  # remove the initial sample

        return pxt_table

    def _distributed_describe(
        self,
        pxt_table: Table,
        dataset: Dataset,
    ) -> Table:
        raise NotImplementedError("Distributed description is not implemented yet.")

    def _sequential_describe(
        self,
        pxt_table: Table,
        dataset: Dataset,
    ) -> Table:
        assert self.batch_size is not None
        assert self.num_workers is not None
        dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers if self.num_workers is not None else 0,
            collate_fn=self.collate_fn,
        )
        for batch in dataloader:
            batch["features"] = self.extractor.extract_and_fuse(
                batch["data"].to(device=self.device)
            )
            if "idx" not in batch:
                batch["idx"] = [idx for idx in range(len(batch["data"]))]
            pxt_table.insert(
                [
                    {
                        key: (
                            (
                                value[sample_idx].item()
                                if value.ndim == 1
                                else value[sample_idx].detach().cpu().numpy()
                            )
                            if isinstance(value, torch.Tensor)
                            else value[sample_idx]
                        )
                        for key, value in batch.items()
                    }
                    for sample_idx in range(len(batch["data"]))
                ]
            )

        return pxt_table
